#line 1275 "xc.nw"
#include "XCGraph.h"
#line 391 "xc.nw"
XCGraph::VIter  XCGraph::inode(const void* p) const
{
   Node  candid(p);
   VIter where = _vertices.find(&candid);
   return where;
} 
#line 400 "xc.nw"
const XCGraph::Node*  XCGraph::node(const void* p) const
{
   VIter  where = inode(p);
   VIter  vEnd = _vertices.end();
   const Node*  pn = (where != vEnd ? where.node() : 0);
   return pn;
} 
#line 417 "xc.nw"
bool XCGraph::eq(const XCGraph& g0, const XCGraph& g1)
{
   bool eq = (g0._vertices == g1._vertices);
   return eq;
}
#line 638 "xc.nw"
XCGraph::XCGraph(const XCGraph& x)
{
   copyFrom(x);
}
#line 649 "xc.nw"
XCGraph& XCGraph::operator=(const XCGraph& rhs)
{
   if (this != &rhs)
   {
      clear();
      copyFrom(rhs);
   }
   return *this;
}
#line 667 "xc.nw"
void XCGraph::clear()
{
   for (PNodeSet::iterator  i = _vertices.begin(), e = _vertices.end();
        i != e;  ++i)
   {
      Node* p = *i;
      delete p;
   }
   _vertices.clear();
}
#line 691 "xc.nw"
void  XCGraph::copyFrom(const XCGraph& x)
{
   MapN2N  n2n;
   PNodeSet::iterator  hint = _vertices.end();
   for (PNodeSet::const_iterator 
        i = x._vertices.begin(), e = x._vertices.end(); i != e;  ++i)
   {
      const Node* oldNode = *i;
      Node*       newNode = new Node(oldNode->ptr()); // just the pointer
      n2n[oldNode] = newNode;
      hint = _vertices.insert(hint, newNode);
   }
   for (PNodeSet::const_iterator 
        i = x._vertices.begin(), e = x._vertices.end(); i != e;  ++i)
   {
      const Node* oldNode = *i;
      Node*       newNode = n2n[oldNode];
      copyNodeSet(newNode->adjacents(),  oldNode->adjacents(), n2n);
      copyNodeSet(newNode->xAdjacents(), oldNode->xAdjacents(), n2n);
   }
}
#line 716 "xc.nw"
void XCGraph::copyNodeSet
(
   XCGraph::PNodeSet&        nsTo,
   const XCGraph::PNodeSet&  nsFrom,
   XCGraph::MapN2N&          n2n
)
{
   PNodeSet::iterator  hint = nsTo.end();
   for (PNodeSet::const_iterator  i = nsFrom.begin(), e = nsFrom.end();
        i != e;  ++i)
   {
      const Node*       oldNode = *i;
      MapN2N::iterator  where = n2n.find(oldNode); // must succeed
      Node*             newNode = (*where).second;
      hint = nsTo.insert(hint, newNode);
   }
}
#line 741 "xc.nw"
bool operator==
(
   const XCGraph::PNodeSet&  ns0, 
   const XCGraph::PNodeSet&  ns1
)
{
   bool  eq = eqPtrs(ns0, ns1);
   if (eq)
   {
      for (XCGraph::PNodeSet::const_iterator
           i = ns0.begin(), e = ns0.end();  eq && (i != e);  ++i)
      {
         const XCGraph::Node*  pNode0 = *i;
         XCGraph::PNodeSet::const_iterator  where = 
            ns1.find(const_cast<XCGraph::Node*>(pNode0));
         eq = (where != ns1.end());
         if (eq)
         {
            const XCGraph::Node*  pNode1 = *where;
            eq = eqPtrs(pNode0->adjacents(), pNode1->adjacents());
            eq = eq && eqPtrs(pNode0->xAdjacents(), pNode1->xAdjacents());
         }
      }
   }
   return eq;
}
#line 770 "xc.nw"
bool  eqPtrs(const XCGraph::PNodeSet&  ns0, const XCGraph::PNodeSet&  ns1)
{
   bool  eq = (ns0.size() == ns1.size());
   if (eq)
   {
      std::set<const void*>  vps[2];
      for (unsigned  si = 0;  eq && (si != 2);  ++si)
      {
         const XCGraph::PNodeSet*      pns = (si ? &ns1 : &ns0);
         std::set<const void*>&            vpsi = vps[si];
          std::set<const void*>::iterator  hint = vpsi.end();
         for (XCGraph::PNodeSet::const_iterator 
                i = pns->begin(), e = pns->end();  eq && (i != e);  ++i)
         {
            hint = vpsi.insert(hint, (*i)->ptr());
         }
      }
      eq = eq && (vps[0] == vps[1]);
   }
   return eq;
}
#line 800 "xc.nw"
void XCGraph::addEdge(const void* from, const void* to)
{
   Node                ndFrom(from);
   PNodeSet::iterator  itrFrom = _vertices.find(&ndFrom);
   if (itrFrom == _vertices.end())
   {
      itrFrom = _vertices.insert(itrFrom, new Node(from));
   }
   Node                ndTo(to);
   PNodeSet::iterator  itrTo = _vertices.find(&ndTo);
   if (itrTo == _vertices.end())
   {
      itrTo = _vertices.insert(itrTo, new Node(to));
   }
   (*itrFrom)->addAdjacent(*itrTo);
}
#line 826 "xc.nw"
void XCGraph::FloydWarshall_xClosure()
{
   
#line 836 "xc.nw"
unsigned  n = _vertices.size();
map<const Node*, unsigned>  v2i;
map<unsigned, Node*>  i2v;
{
   unsigned                    k = 0;
   for (PNodeSet::iterator
        vi = _vertices.begin(), e = _vertices.end();  vi != e;  ++vi, ++k)
   {
      Node* p = *vi;
      v2i[p] = k;
      i2v[k] = p;
   }
}
#line 853 "xc.nw"
const std::vector<bool>     falseRow((size_t)n, false);
std::vector<vector<bool> >  T((size_t)n, falseRow);
#line 859 "xc.nw"
for (PNodeSet::const_iterator
     vi = _vertices.begin(), er = _vertices.end();  vi != er;  ++vi)
{
   const Node*      pi = *vi;
   unsigned         i = v2i[pi];
   const PNodeSet&  adjSet = pi->adjacents();
   for (XCGraph::PNodeSet::const_iterator
        vj = adjSet.begin(), ec = adjSet.end();  vj != ec;  ++vj)
   {
      const Node*      pj = *vj;
      unsigned         j = v2i[pj];
      T[i][j] = true;
   }
}
#line 878 "xc.nw"
for (unsigned k = 0;  k != n;  ++k)
{
   for (unsigned i = 0;  i != n;  ++i)
   {
      for (unsigned j = 0;  j != n;  ++j)
      {
         T[i][j] = T[i][j] || (T[i][k] && T[k][j]);
      }
   }
}
#line 892 "xc.nw"
for (unsigned i = 0;  i != n;  ++i)
{
   Node* pi = i2v[i];
   for (unsigned j = 0;  j != n;  ++j)
   {
      if (T[i][j])
      {
         Node* pj = i2v[j];
         pi->xAdjacents().insert(pj);
      }
   }
}
#line 829 "xc.nw"
}
#line 921 "xc.nw"
void XCGraph::DFS_xClosure()
{
   _stack.reserve(_vertices.size());
   _stack.clear();
   for (PNodeSet::iterator i = _vertices.begin(), e = _vertices.end();
        i != e;  ++i)
   {
      Node*  pNode = *i;
      Node&  nd = *pNode;
      nd.color(white);
      nd.xAdjacents() = nd.adjacents();
      nd.rep(pNode);
      nd.Comp().insert(pNode);
   }
   for (PNodeSet::iterator i = _vertices.begin(), e = _vertices.end();
        i != e;  ++i)
   {
      Node&  nd = *(*i);
      if (nd.color() == white)
      {
         vxClosure(nd);
      }
   }
}
#line 1082 "xc.nw"
template<class Set>
void add2Set(Set& to, const Set& from)
{
   typename  Set::iterator  hint = to.begin();
   for (typename  Set::const_iterator  i = from.begin(), e = from.end();
        i != e;  ++i)
   {
      hint = to.insert(hint, *i);
   }
}
#line 965 "xc.nw"
void XCGraph::vxClosure(XCGraph::Node& v)
{ 
   v.color(gray);
   v.position(_stack.size());
   _stack.push_back(&v);
   for (PNodeSet::iterator i = v.adjacents().begin(), e = v.adjacents().end();
        i != e;  ++i)
   {
      Node&  w = *(*i);
      Color  cw = w.color();
      switch (cw)
      {
       case white:
         vxClosure(w);
         break;
       case gray:
         grayCycle(v, w);
         break;
       default:;
      }
      if (w.color() == black)
      {
         add2Set(v.xAdjacents(), w.xAdjacents());
      }
   }
   if (v.rep() == &v)
   {
      blackCycle(v);
   }
   _stack.pop_back();
}
#line 1009 "xc.nw"
void XCGraph::grayCycle(XCGraph::Node& v, XCGraph::Node& w)
{
   PNodeSet  A;
   Node*     repw = w.rep();
   Node*     s = _stack.back();
   while (s != repw)
   {
      add2Set(A, s->Comp());
      s->Comp().clear();
      unsigned  sPosition = s->position();
      s = _stack[sPosition - 1]->rep();
   }
   for (PNodeSet::const_iterator  i = A.begin(), e = A.end();  i != e;  ++i)
   {
      (*i)->rep(repw);
   }
   add2Set(repw->Comp(), A);
}
#line 1044 "xc.nw"
void XCGraph::blackCycle(XCGraph::Node& v)
{
   if (v.Comp().size() == 1)
   {
      v.color(black);
   }
   else
   {
      PNodeSet  A;
      for (PNodeSet::const_iterator i = v.Comp().begin(), e = v.Comp().end();
           i != e;  ++i)
      {
         const Node& w = *(*i);
         add2Set(A, w.xAdjacents());
      }
      for (PNodeSet::iterator i = v.Comp().begin(), e = v.Comp().end();
           i != e;  ++i)
      {
         Node& w = *(*i);
         w.xAdjacents() = A;
         w.color(black);
      }
   }
}
#line 1108 "xc.nw"
XCGraph::Color XCGraph::Node::color() const 
{
   Color  c =  (((_position == white) || (_position == black))
                ? (Color)_position
                : gray);
   return c;
}
