#line 237 "xc.nw"
#ifndef xcgrpAh_H
#define xcgrpAh_H

#line 1267 "xc.nw"
#include <functional>
#include <string>
#include <map>
#include <set>
#include <vector>
#line 257 "xc.nw"
class XCGraph
{
 public:
   
#line 285 "xc.nw"
enum Color {white = -1, gray = -2, black = -3};
#line 302 "xc.nw"
class Node;
#line 343 "xc.nw"
class NodeLessThan
{
 public:
   NodeLessThan() {}
   virtual ~NodeLessThan() {}
   bool operator()(const Node& n0, const Node& n1) const 
   { return n0.ptr() < n1.ptr(); }
};
#line 354 "xc.nw"
class PNodeLessThan
{
 public:
   PNodeLessThan() {}
   virtual ~PNodeLessThan() {}
   bool operator()(const Node* n0, const Node* n1) const 
   { return n0->ptr() < n1->ptr(); }
};
#line 366 "xc.nw"
// typedef std::set<Node,  NodeLessThan>   NodeSet;
typedef std::set<Node*, PNodeLessThan>  PNodeSet;
typedef std::map<const Node*, Node*>          MapN2N;
#line 304 "xc.nw"
class Node
{
 public:
   typedef XCGraph::Color  Color;
   Node(const void* p = 0) : _ptr(p) {}
   virtual ~Node() {}
   // virtual Node*    clone() const {return new Node(*this);}
   const void*      ptr() const {return _ptr;}
   Color            color() const;
   unsigned         position() const {return _position;};
   void             position(int i) {_position = i;};
   
#line 334 "xc.nw"
unsigned   nAdjacents()  const {return _adjacents.size();}
unsigned   nXAdjacents() const {return _xAdjacents.size();}
#line 1104 "xc.nw"
void color(Color clr) {_position = clr;}
#line 1118 "xc.nw"
Node*  rep() const   {return _rep;}
void   rep(Node* pn) {_rep = pn;}
#line 1126 "xc.nw"
const PNodeSet&  Comp() const {return _Comp;}
      PNodeSet&  Comp()       {return _Comp;}
#line 316 "xc.nw"
   const PNodeSet&  adjacents() const {return _adjacents;}
         PNodeSet&  adjacents()       {return _adjacents;}
   const PNodeSet&  xAdjacents() const {return _xAdjacents;}
         PNodeSet&  xAdjacents()       {return _xAdjacents;}
   void             addAdjacent(Node* to) {_adjacents.insert(to);}
 private:
   Node(const Node&);            // N.A.
   Node& operator=(const Node&); // N.A.
   const void*  _ptr;
   int          _position;
   PNodeSet     _adjacents;
   PNodeSet     _xAdjacents;
   
#line 1122 "xc.nw"
Node*  _rep;
#line 1131 "xc.nw"
PNodeSet  _Comp;
#line 329 "xc.nw"
};
#line 436 "xc.nw"
class VIter
{
 public: 
   VIter() {}
   VIter(PNodeSet::const_iterator itr) : _sItr(itr) {}
   virtual ~VIter() {}
   static bool eq(const VIter& vi0, const VIter& vi1) 
      {return vi0._sItr == vi1._sItr;}
   const void*  operator*() const {return (*_sItr)->ptr();}
   const Node*  node()      const {return *_sItr;}
   VIter&       operator++()      {++_sItr; return *this;}
   VIter        operator++(int)   {VIter tmp(*this); ++_sItr; return tmp;}
 public:
   PNodeSet::const_iterator  _sItr;
};
#line 261 "xc.nw"
   XCGraph() {}
   XCGraph(const XCGraph&);
   XCGraph& operator=(const XCGraph&);
   virtual ~XCGraph() {clear();}
   void         addEdge(const void* from, const void* to);
   void         FloydWarshall_xClosure();
   void         DFS_xClosure();
   unsigned     nV() const {return _vertices.size();}
   VIter        inode(const void* p) const;
   const Node*  node(const void* p) const;
   VIter        vBegin() const {return _vertices.begin();}
   VIter        vEnd()   const {return _vertices.end();}
   static bool  eq(const XCGraph&, const XCGraph&);
 private:
   static void  copyNodeSet(PNodeSet&, const PNodeSet&, MapN2N&);
   void         clear();
   void         copyFrom(const XCGraph&);
   
#line 961 "xc.nw"
void vxClosure(Node& v);
#line 1006 "xc.nw"
void grayCycle(Node& v, Node& w);
#line 1041 "xc.nw"
void blackCycle(Node& v);
#line 279 "xc.nw"
   
#line 384 "xc.nw"
PNodeSet  _vertices;
#line 948 "xc.nw"
std::vector<Node*>  _stack;
#line 280 "xc.nw"
};
#line 289 "xc.nw"
inline bool operator==(const XCGraph& g0, const XCGraph& g1)
{ return XCGraph::eq(g0, g1); }
inline bool operator!=(const XCGraph& g0, const XCGraph& g1)
{ return !(g0 == g1); }
#line 455 "xc.nw"
inline bool operator==(const XCGraph::VIter& vi0, 
                       const XCGraph::VIter& vi1)
{ return XCGraph::VIter::eq(vi0, vi1); }
inline bool operator!=(const XCGraph::VIter& vi0, 
                       const XCGraph::VIter& vi1)
{ return !(vi0 == vi1); }
#line 521 "xc.nw"
class _XCTGraphTIterBase
{
 public:
   _XCTGraphTIterBase() {}
   _XCTGraphTIterBase(const XCGraph::VIter& vi) : _i(vi) {}
   virtual ~_XCTGraphTIterBase() {}
   static bool eq(const _XCTGraphTIterBase& i0, const _XCTGraphTIterBase& i1)
               {return (i0._i == i1._i);}
 protected:
   XCGraph::VIter  _i;
};
#line 560 "xc.nw"
template<class T>
class XCTGraph
{
 public:
   XCTGraph() {}
   virtual ~XCTGraph() {}
   void         addEdge(const T* from, const T* to) {_v.addEdge(from, to);}
   void         FloydWarshall_xClosure()   {_v.FloydWarshall_xClosure();}
   void         DFS_xClosure()             {_v.DFS_xClosure();}
   unsigned     nV() const                 {return _v.nV();}
   
#line 484 "xc.nw"
class TIter; // forward
class TNode
{
 public:
   typedef XCTGraph<T>::TNode  self;
   typedef XCTGraph<T>::TIter  TIter;
   TNode() {}
   TNode(const XCGraph::VIter& i) : _i(i) {}
   virtual ~TNode() {}
   const T*     ptr() const {return (const T*)(*_i)->ptr();}
   static bool  eq(const self& ni0, const self& ni1) 
                {return ni0._i == ni1._i;}
   XCGraph::VIter     vadjacentsBegin()  
                      const {return node().adjacents().begin();}
   unsigned  nAdjacents()      const {return adjs().size();}
   TIter     adjacentsBegin()  const {return TIter(adjs().begin());}
   TIter     adjacentsEnd()    const {return TIter(adjs().end());}
   unsigned  nXAdjacents()     const {return xAdjs().size();}
   TIter     xAdjacentsBegin() const {return TIter(xAdjs().begin());}
   TIter     xAdjacentsEnd()   const {return TIter(xAdjs().end());}
 private:
   const XCGraph::Node&  node() const {return *(_i.node());}
   const XCGraph::PNodeSet&    adjs()  const {return node().adjacents();}
   const XCGraph::PNodeSet&    xAdjs() const {return node().xAdjacents();}
   XCGraph::VIter  _i;
};
#line 571 "xc.nw"
   
#line 536 "xc.nw"
class TIter : public _XCTGraphTIterBase
{
 public:
   typedef XCTGraph<T>::TIter  self;
   typedef XCTGraph<T>::TNode  TNode;
   TIter() {}
   TIter(const XCGraph::VIter& vi) : _XCTGraphTIterBase(vi) {}
   virtual ~TIter() {}
   static bool eq(const self& vi0, const self& vi1) 
      {return TNode::eq(vi0._tnode, vi1._tnode);}
   const T*  ptr()            const {return (const T*)(*_i);}
   // const XCTGraph<T>::TNode*  operator->() const;
   TNode                 operator*() const {return TNode(_i);}
   self& operator++()    {++_i; return *this;}
   self  operator++(int) {self tmp(*this); ++_i; return tmp;}
};
#line 572 "xc.nw"
   TIter        node(const T* t) const {return _v.inode(t);}
   TIter        vBegin() const         {return _v.vBegin();}
   TIter        vEnd() const           {return _v.vEnd();}
   static bool  eq(const XCTGraph<T>& g0, const XCTGraph<T>& g1)
                {return XCGraph::eq(g0._v, g1._v);}
 public:
   XCGraph  _v;  
};

#line 373 "xc.nw"
bool eqPtrs(const XCGraph::PNodeSet&, const XCGraph::PNodeSet&);
bool operator==(const XCGraph::PNodeSet&, const XCGraph::PNodeSet&);
inline bool 
operator!=(const XCGraph::PNodeSet& ns0, const XCGraph::PNodeSet& ns1)
{ return !(ns0 == ns1); }
#line 592 "xc.nw"
template<class T>
inline bool operator==(const XCTGraph<T>& g0, const XCTGraph<T>& g1)
{ return XCTGraph<T>::eq(g0, g1); }
template<class T>
inline bool operator!=(const XCTGraph<T>& g0, const XCTGraph<T>& g1)
{ return !(g0 == g1); }
#line 604 "xc.nw"
template<class T>
inline bool operator==(const XCTGraph<T>::TIter& i0, 
                       const XCTGraph<T>::TIter& i1)
{ return XCTGraph<T>::TIter::eq(i0, i1); }
template<class T>
inline bool operator!=(const XCTGraph<T>::TIter& i0, 
                       const XCTGraph<T>::TIter& i1)
{ return !(i0 == i1); }
#line 616 "xc.nw"
inline bool operator==(const _XCTGraphTIterBase& i0, 
                       const _XCTGraphTIterBase& i1)
{ return _XCTGraphTIterBase::eq(i0, i1); }
inline bool operator!=(const _XCTGraphTIterBase& i0, 
                       const _XCTGraphTIterBase& i1)
{ return !(i0 == i1); }
#line 244 "xc.nw"
#endif /* xcgrpAh_H */
