#line 1323 "mcdep.nw"
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <algorithm>
#include <iterator>
#include <list>
#include <map>
#include <set>
#include <utility>
#include <vector>
#include "tools.h"
#include "FileLines.h"
#include "XCGraph.h"
#line 432 "mcdep.nw"
#include "IncComponent.h"
#line 115 "mcdep.nw"
class Unit;
class UnitLessThan : binary_function<Unit*, Unit*, bool>
{
 public:
   UnitLessThan() {}
   bool operator()(const Unit* pu0, const Unit* pu1);
};
#line 82 "mcdep.nw"
class McDep;
class IncInfo;
class ostream;
class FileLines;
class Unit
{
 public:
   typedef XCTGraph<Unit>  XCUGraph;
   Unit(const string& fn, const string& mkSymNm) :
      _fileName(fn), _makeSymName(mkSymNm), _completed(false) {}
   void  processIncludes
#line 724 "mcdep.nw"
(
   McDep&                  mcDep, 
   const vector<IncInfo>&  incs
)
#line 92 "mcdep.nw"
							  ;
   bool  completed() const { return _completed; }
   typedef set<const Unit*, UnitLessThan>  DepSet;
   static bool  lt(const Unit& u0, const Unit u1) 
      { return u0._fileName < u1._fileName; }
   unsigned        depOut(ostream&, const XCUGraph&, bool doSort) const;
   const string&   symbolicName() const {return _makeSymName;}
 private:
   // Unit() {}
   
#line 754 "mcdep.nw"
void doInclude
#line 757 "mcdep.nw"
(McDep&               mcDep, 
 const IncComponent*  localComp, 
 const string&        header)
#line 754 "mcdep.nw"
					     ;
#line 102 "mcdep.nw"
   
#line 107 "mcdep.nw"
string   _fileName;
string   _makeSymName;      
bool     _completed;
#line 103 "mcdep.nw"
};
#line 126 "mcdep.nw"
inline  bool UnitLessThan::operator()(const Unit* pu0, const Unit* pu1)
{ return Unit::lt(*pu0, *pu1); }
#line 331 "mcdep.nw"
static const  string  noOriginDir(" ");
#line 338 "mcdep.nw"
typedef pair<string, string>  head_reldir;
typedef map<head_reldir, Unit const*> IncludeKnowledge;
#line 681 "mcdep.nw"
#define _VERB(_v, verbFlag, streamContent) \
 if ((_v) & (verbFlag)) {cerr << streamContent << endl;}
#define VERB(verbFlag, streamContent) _VERB(_verbose, verbFlag, streamContent)
#define UVERB(verbFlag, streamContent) \
   _VERB(mcDep.verbose(), McDep::verbFlag, streamContent)
#line 937 "mcdep.nw"
class IncInfo
{
 public:
   IncInfo(const string& h = "", bool r = false) : _header(h), _relative(r) {}
   virtual ~IncInfo() {}
   const string& header() const {return _header;}
   bool          relative() const {return _relative;}
 private:
   string _header;
   bool   _relative;
};
#line 446 "mcdep.nw"
class McDep
{
 public:
   
#line 463 "mcdep.nw"
typedef Unit::XCUGraph  XCUGraph;
#line 662 "mcdep.nw"
enum 
{
   verbStat     =   1,
   verbOpen     =   2,
   verbOpenFail =   4,
   verbProcess  =   8,
   verbClose    =  16,
   verbInc      =  32,
   verbSearch   =  64,
   verbRoot     = 128,
   verbDefault  = verbStat
};
#line 450 "mcdep.nw"
   McDep(int argc, char const* const* argv);
   ~McDep() {
#line 599 "mcdep.nw"
if (_pOut != &cout)
{
   delete _pOut;
}
#line 1038 "mcdep.nw"
for (list<Unit*>::iterator  i = _unitList.begin(), e = _unitList.end();  
     i != e;  ++i)
{
   Unit*  pu = (*i);
   delete pu;
}
#line 451 "mcdep.nw"
			   }
   bool  ok() const {return _ok;}
   bool  run();
   
#line 347 "mcdep.nw"
const Unit*  incKnown(const head_reldir& key) const
{
   IncludeKnowledge::const_iterator  w = _incKnowledge.find(key);
   Unit const*  pu = (w != _incKnowledge.end() ? (*w).second : 0);
   return pu;
}
#line 356 "mcdep.nw"
void  incTeach(const head_reldir& key, const Unit* pu) 
{
   _incKnowledge[key] = pu;
}
#line 373 "mcdep.nw"
long int findHeader
#line 376 "mcdep.nw"
(const IncComponent*  localComp, 
 const string&        fileTailName,
 string&              fileName,
 string&              fileSymbolicName,
 string&              locDirSym)
#line 373 "mcdep.nw"
					   ;
#line 689 "mcdep.nw"
unsigned long  verbose() const {return _verbose;}
#line 794 "mcdep.nw"
void graphEdge(Unit const* from, Unit const* to) { _uGraph.addEdge(from, to); }
#line 954 "mcdep.nw"
bool  McDep::parse(vector<IncInfo>& includes, const string& fn, int preSz = -1);
#line 1175 "mcdep.nw"
bool  statistics() const;
#line 455 "mcdep.nw"
 private:
   
#line 565 "mcdep.nw"
bool  ctorFinal(const char* objFmt, const char* outFmt)
{
   bool   ok = fmt2fices(objFmt, _objPrefix, _objSuffix);
   _multiOut = fmt2fices(outFmt, _outPrefix, _outSuffix);
   if (!_multiOut)
   {
      _pOut = &cout;
      if (strcmp(outFmt, "-") != 0)
      {
         _pOut = new ofstream(outFmt);
      }
   }
   return ok;
}
#line 609 "mcdep.nw"
bool  fmt2fices(const char* fmt, string& prefix, string& suffix)
{
   const char*  persent = strstr(fmt, "%s");
   if (persent != 0)
   {
      prefix = string(fmt, persent - fmt);
      suffix = string(persent + 2);
   }
   return (persent != 0);
}
#line 626 "mcdep.nw"
static void usage(const char* argv0)
{
   cerr << 
   "McDep - Make C Make-Dependencies." <<     endl << 
   "Usage: " <<                               endl << 
   "   " << argv0 <<                          endl <<
   "      [-Ipath1 ... -Ipathm]" <<           endl << 
   "      [-Jpath1 sym1 ... -Jpathn symn]" << endl << 
   "      [-obj <objFmt>]" <<                 endl <<
   "      [-o <out>]" <<                      endl <<
   "      [-sort]" <<                         endl <<
   "      [-verbose <verbose flags]" <<       endl << 
   "      <source roots files list>" <<       endl << 
                                              endl <<
   "Where:" <<                                endl <<
   "  symi   - symbol to replace the path in dependency record" <<      endl <<
   "  objFmt - format symbol for the target object. Default \"%s.o\"" << endl <<
   "  out    - Output. With \"%s\" for multi pattern named files." <<    endl <<
   "  verbose flags - combination of bits:" <<                           endl <<
   "     total statistics:  " << setw(3) << verbStat     << endl <<
   "     file open:         " << setw(3) << verbOpen     << endl <<
   "     open failure:      " << setw(3) << verbOpenFail << endl <<
   "     Process file:      " << setw(3) << verbProcess  << endl <<
   "     file close:        " << setw(3) << verbClose    << endl <<
   "     inclusion:         " << setw(3) << verbInc      << endl <<
   "     search:            " << setw(3) << verbSearch   << endl <<
   "     top files:         " << setw(3) << verbRoot     << endl <<
   "  default value:        " << setw(3) << verbDefault  << endl;
}           
#line 802 "mcdep.nw"
bool  isInclude(IncInfo&, const char* line);
#line 897 "mcdep.nw"
bool  McDep::parseRoots(std::vector<Unit*>& units);
#line 997 "mcdep.nw"
void  McDep::output(const std::vector<Unit*>& units);
#line 1197 "mcdep.nw"
enum {
   stOpen,
   stOpenFail,
   stSearch,
   stLines,
   stIncRel,
   stIncNonRel,
   stDepend,
   stat_N
};
#line 1209 "mcdep.nw"
unsigned  _statCounters[stat_N];
#line 457 "mcdep.nw"
   
#line 343 "mcdep.nw"
IncludeKnowledge  _incKnowledge;
#line 536 "mcdep.nw"
bool                  _ok;
vector<IncComponent>  _includePath;
int                   _nRoots;
char const* const*    _roots;
bool                  _sort;
unsigned long         _verbose;
XCUGraph              _uGraph;
#line 587 "mcdep.nw"
string    _objPrefix;
string    _objSuffix;
string    _outPrefix;
string    _outSuffix;
bool      _multiOut;
ostream*  _pOut;
#line 1033 "mcdep.nw"
list<Unit*>   _unitList;
#line 458 "mcdep.nw"
};
#line 153 "mcdep.nw"
unsigned  Unit::depOut(ostream& os, const XCUGraph& uGraph, bool doSort) const
{
   unsigned  nDeps = 0;
   XCUGraph::TIter  u = uGraph.node(this);
   if (u != uGraph.vEnd())
   {
      XCUGraph::TNode           node = *u;
      nDeps = node.nXAdjacents();
      vector<string>            deps(nDeps, string(""));
      vector<string>::iterator  depsIter = deps.begin();
      for (XCUGraph::TIter  
             i = node.xAdjacentsBegin(), e = node.xAdjacentsEnd(); 
           i != e;  ++i, ++depsIter)
      {
         const Unit*  uDep = i.ptr();
         *depsIter = uDep->_makeSymName;
      }
      if (doSort)
      {
         sort(deps.begin(), deps.end());
      }
      for (vector<string>::const_iterator i = deps.begin(), e = deps.end();  
           i != e;  ++i)
      {
         os << " \\" << endl <<
               "\t" << *i;
      }
   }
   os << endl << endl;
   return nDeps;
}
#line 730 "mcdep.nw"
void  Unit::processIncludes
#line 724 "mcdep.nw"
(
   McDep&                  mcDep, 
   const vector<IncInfo>&  incs
)
#line 730 "mcdep.nw"
							      
{
   UVERB(verbProcess, "( begin processing: " << _fileName);
   const string  localDir = uptoDir(_fileName);
   const string  symLocDir = uptoDir(_makeSymName);
   IncComponent  localComp(localDir, symLocDir);
   for (vector<IncInfo>::const_iterator  i = incs.begin(), e = incs.end();
        i != e;  ++i)
   {
      const IncInfo& ii = *i;
      doInclude(mcDep, (ii.relative() ? &localComp : 0), ii.header());
   }
   UVERB(verbProcess, ") end processing: " << _fileName);
   _completed = true;
}
#line 762 "mcdep.nw"
void Unit::doInclude
#line 757 "mcdep.nw"
(McDep&               mcDep, 
 const IncComponent*  localComp, 
 const string&        header)
#line 763 "mcdep.nw"
{
   static const Unit  unitNotFound("", ""); // just to serve as an address tag
   head_reldir  key = 
      make_pair(header, (localComp ? localComp->directory() : noOriginDir));
   const Unit*  pu = mcDep.incKnown(key);
   if (pu == 0)
   {
      string    hdrFileName, symName, symDir;
      long int  hdrSize = 
         mcDep.findHeader(localComp, header, hdrFileName, symName, symDir);
      mcDep.incTeach(key, &unitNotFound); // pessimistic default
      if (hdrSize != -1)
      {
         vector<IncInfo>  incs;
         if (mcDep.parse(incs, hdrFileName, hdrSize)) 
         {
            Unit*  hdrUnit = new Unit(hdrFileName, symName);
            pu = hdrUnit;
            mcDep.incTeach(key, pu); // must know before parse
            hdrUnit->processIncludes(mcDep, incs);
         }
      }
   }
   if ((pu != 0) && (pu != &unitNotFound))
   {
      mcDep.graphEdge(this, pu);
   }
}
#line 383 "mcdep.nw"
long int McDep::findHeader
#line 376 "mcdep.nw"
(const IncComponent*  localComp, 
 const string&        fileTailName,
 string&              fileName,
 string&              fileSymbolicName,
 string&              locDirSym)
#line 384 "mcdep.nw"
{
   int  statSize = -1;
   vector<IncComponent>::iterator  
           i = _includePath.begin(), // start with local
           e = _includePath.end();  
   if (localComp)
   {
      *i = *localComp;
   }
   else
   {
      ++i;
   }
   for (; (statSize == -1) && (i != e);  ++i)
   {
      const IncComponent&  incComp = *i;
      ++_statCounters[stSearch];
      statSize = incComp.hasMatch(fileTailName, fileName, fileSymbolicName, 
                                  locDirSym, ((_verbose & verbSearch) != 0));
   }
   if (_verbose & verbSearch)
   {
      if (statSize == -1)
      {
         cerr << fileTailName << " not found" << endl;
      }
      else
      {
         cerr << fileTailName << " found in \"" << fileName << 
                 "\" Make-symbol=\"" << fileSymbolicName << 
                 "\" size=" << statSize << endl;
      }
   }

   return statSize;
}
#line 477 "mcdep.nw"
McDep::McDep(int argc, char const* const* argv) :
   _ok(true),
   _sort(false),
   _verbose(verbDefault),
   _pOut(0)
{
   fill_n(_statCounters, (size_t)stat_N, 0);
   int          i;
   const char*  objFmt = "%s.o";
   const char*  outFmt = "-";
   _includePath.push_back(IncComponent());
   for (i = 1;  _ok && (i < argc - 1) && (argv[i][0] == '-');  )
   {
      char const*     arg  = argv[i];
      const string    sarg(arg);
      char const*     arg1 = argv[i + 1];
      const string    sarg1(arg1);
      char            copt = arg[1];
      if (copt == 'I' || copt == 'J')
      {                // skip  "-I"
         const string  iDir(arg + 2);  i += 1;
         string        sym(iDir);
         if (copt == 'J')
         {
            sym = sarg1;               i += 1;
         }
         _includePath.push_back(IncComponent(iDir, sym));
      }
      else if (sarg == string("-obj"))
      {
         objFmt = arg1;                i += 2;
      }
      else if (sarg == string("-o"))
      {
         outFmt = arg1;                i += 2;
      }
      else  if (sarg == string("-sort"))
      {
         _sort = true;                 i += 1;
      }
      else  if (sarg == string("-verbose"))
      {
         _verbose = atol(arg1);        i += 2;
      }
      else
      {
         _ok = false;
         cerr << argv[0] << ": Unrecognized option: " << sarg << endl;
      }
   }
   _nRoots = argc - i;
   _roots  = argv + i;
   _ok = _ok && (_nRoots > 0) && ctorFinal(objFmt, outFmt);
   if (!_ok) {usage(argv[0]);}
}
#line 805 "mcdep.nw"
bool  McDep::isInclude(IncInfo& incInfo, const char* line)
{
   bool  isInc = false;
   line = skipWhiteSpaces(line);
   if (line[0] == '#')
   {
      line = skipWhiteSpaces(line + 1);
      static const char*     cs_include = "include";
      static const unsigned  includeLen  = strlen(cs_include);
      isInc = (strncmp(line, cs_include, includeLen) == 0);
      if (isInc) 
      {
         const char* incArg = skipWhiteSpaces(line + includeLen);
         char  c = *incArg;
         bool  relative = (c == '\"');
         isInc = relative || (c == '<');
         if (isInc) 
         {
            char         delimeter = (c == '<' ? '>' : '\"');
            const char*  endHeader = strchr(++incArg, delimeter);
            isInc = (endHeader != 0);
            if (isInc)
            {
               incInfo = IncInfo(string(incArg, endHeader - incArg), relative);
            }
         }
      }
   }
   return isInc;
}
#line 875 "mcdep.nw"
bool  McDep::run()
{
   vector<Unit*>  units(_nRoots, (Unit*)0);
   if (parseRoots(units))
   {
      _uGraph.DFS_xClosure();
      output(units);
   }
   
   for (vector<Unit*>::iterator i = units.begin(), e = units.end(); i != e; ++i)
   {
      delete *i;
   }
   return _ok;
}
#line 900 "mcdep.nw"
bool  McDep::parseRoots(std::vector<Unit*>& units)
{
   for (int  ri = 0;  _ok && (ri < _nRoots);  ++ri)
   {
      const char*      root = _roots[ri];
      VERB(verbRoot, "Processing root file: " << root); 
      vector<IncInfo>  incs;
      if (parse(incs, root)) 
      {
         const string   sRoot(root);
         units[ri] = new Unit(root, sRoot);
         Unit&          unit = *(units[ri]);
         unit.processIncludes(*this, incs);
      }
      else
      {
         cerr << "Failed to open root file: " << root << endl;
         _ok = false;
      }
   }
   return _ok;
}
#line 957 "mcdep.nw"
bool  McDep::parse(vector<IncInfo>& incs, const string& fn, int preSize)
{
   VERB(verbOpen, "Opening: " << fn); 
   _statCounters[stSearch] += (preSize == -1 ? 1 : 0);
   FileLines  fileLines(fn.c_str(), preSize);
   bool       ok = fileLines.ok();
   if (ok)
   {
      ++_statCounters[stOpen];
      _statCounters[stLines] += fileLines.nLines();
      incs.clear();
      for (unsigned  li = 0, li1 = 1, nL = fileLines.nLines();  
           li != nL;  li = li1++)
      {
         IncInfo      incInfo;
         const char*  line = fileLines.line(li);
         if (isInclude(incInfo, line))
         {
            VERB(verbInc, fn << "[" << li1 << "]: " << line);
            incs.push_back(incInfo);
            ++_statCounters[incInfo.relative() ? stIncRel : stIncNonRel];
         }
      }
      VERB(verbClose, "Closing: " << fn); 
   }
   else
   {
      VERB(verbOpenFail, "Failed to Open: " << fn); 
      ++_statCounters[stOpenFail];
   }
   return ok;
} 
#line 1000 "mcdep.nw"
void  McDep::output(const std::vector<Unit*>& units)
{
   for (vector<Unit*>::const_iterator i = units.begin(), e = units.end(); 
        i != e; ++i)
   {
      const Unit&  unit = *(*i);
      const string&  sRoot = unit.symbolicName();
      ofstream*      pUnitOut = 0;
      if (_multiOut)
      {
         string  depOutName = _outPrefix + uptoSuffix(sRoot) + _outSuffix;
         pUnitOut = new ofstream(depOutName.c_str());
      }
      string  target = _objPrefix + uptoSuffix(sRoot) + _objSuffix;
      *_pOut << target << ": " << sRoot;
      unsigned  nDeps = unit.depOut(*_pOut, _uGraph, _sort);
      _statCounters[stDepend] += nDeps;
      if (pUnitOut)
      {
         delete pUnitOut;
         _pOut = 0;
      }
   }
} 
#line 1178 "mcdep.nw"
bool  McDep::statistics() const
{
   if ((_verbose & verbStat) != 0)
   {
      cerr << "Statistics:" << endl <<
     "  Files opened: " << setw(9) << _statCounters[stOpen]      << endl <<
     "  Open Failures:" << setw(9) << _statCounters[stOpenFail]  << endl <<
     "  Searches:     " << setw(9) << _statCounters[stSearch]    << endl <<
     "  Lines read:   " << setw(9) << _statCounters[stLines]     << endl <<
   "  #includes \"\": " << setw(9) << _statCounters[stIncRel]    << endl <<
     "  #includes <>: " << setw(9) << _statCounters[stIncNonRel] << endl<<
     "  Dependencies: " << setw(9) << _statCounters[stDepend]    << endl;

   }
   return true;
} 
#line 698 "mcdep.nw"
int main(int argc, char** argv)
{
   McDep  mcDep(argc, argv);
   bool   ok = mcDep.ok();
   ok =   ok && mcDep.run() && mcDep.statistics();
   return !ok;
}
